/*
 * ============================================================================
 * GNU Lesser General Public License
 * ============================================================================
 *
 * Square Brackets - Java Array Framework.
 * Copyright (C) 2013  Beatgrid Media B.V.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * Beatgrid Media B.V.
 * Sisalbaan 5-A
 * 2352 AZ Leiderdorp
 * The Netherlands
 *
 * info@squarebrackets.org
 * http://squarebrackets.org
 */
package org.squarebrackets;

import org.squarebrackets.function.ByteConsumer;

import javax.annotation.Nonnull;
import javax.annotation.concurrent.Immutable;

/**
 * Immutable array of byte primitives.
 *
 * <p>Byte arrays are constructed by the static interface methods of this interface. The {@code copyOf} method returns
 * a new array representing a shallow copy of the (native or framework) array argument. The {@code unsafeValueOf}
 * methods wrap a new {@code ByteArray} around a native array. It is imperative that this native array is not 
 * modified outside the scope of {@code ByteArray}, as this would break the immutable contract. The 
 * {@code unsafeValueOf} method is used to prevent unnecessary array copies. It is the caller's responsibility not to 
 * violate the contract.</p>
 *
 * <p>Arrays express characteristics to indicate that, for instance, elements are listed in sorted order, or no
 * duplicate elements exist. These characteristics allow array implementations to optimize their algorithms based
 * on the contents of the array. For example, operations that lookup elements can use a binary search algorithm if
 * the array is marked as being sorted. These optimizations may reduce the time complexity of such operations by
 * several orders of magnitude.</p>
 *
 * <p>The array implementations of the framework attempt to establish these characteristics to accommodate these
 * optimizations at best effort. This means that the array implementation determines which set of characteristics apply
 * to its elements without traversing these elements, in other words, without incurring a performance penalty.
 * Primitive arrays, such as {@code ByteArray}, always report {@code NONNULL}. The set of characteristics of 
 * immutable arrays are constant.</p>
 *
 * @author Leon van Zantvoort
 */
@Immutable
public interface ByteArray extends Array<Byte> {

    /** Empty array. */
    static final ByteArray EMPTY = unsafeValueOf();

    /**
     * Creates an immutable copy of the specified {@code array}. If the specified {@code array} is mutable, changes to
     * the mutable array are not reflected by this immutable copy.
     *
     * <p>The specified {@code array} can be any array of primitive or object numbers. Reconstructing a copy of this
     * array may involve rounding or truncation as implemented by {@link Number#byteValue}. The specified {@code array}
     * must not contain any {@code null} elements.</p>
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code NONNULL};</li>
     *   <li>{@code SORTED} and {@code DISTINCT} if the specified {@code array} contains zero or one element(s);</li>
     *   <li>all characteristics reported by the specified {@code array}, except {@code MUTABLE} and {@code DYNAMIC}.</li>
     * </ul>
     *
     * @param array array to be copied.
     * @return immutable array of byte values.
     * @throws NullPointerException if the specified array is null or if the array contains null elements.
     */
    static ByteArray copyOf(@Nonnull Array<? extends Number> array) {
        return Arrays.newArray(UnsafeArray.byteReadAccess(array), array.offset(), array.length(),
                array.characteristics(), array.hasCharacteristics(MUTABLE) && UnsafeArray.hasDirectAccess(array, Boolean.TYPE));
    }

    /**
     * Creates an immutable array representation of the specified {@code array}. A defensive copy of {@code array} is
     * used as backing array to ensure immutability.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code NONNULL};</li>
     *   <li>{@code SORTED} and {@code DISTINCT} if the specified {@code array} contains zero or one element(s).</li>
     * </ul>
     *
     * @param array array to be copied.
     * @return immutable array of byte values.
     */
    static ByteArray copyOf(@Nonnull byte... array) {
        return copyOf(array, 0, array.length);
    }

    /**
     * <p>Creates an immutable array representation of the specified {@code array}. A defensive copy of {@code array} is
     * used as backing array to ensure immutability.</p>
     *
     * <p>Note that the backing array will have the same length as {@code array}.</p>
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code NONNULL};</li>
     *   <li>{@code SORTED} and {@code DISTINCT} if the specified {@code length} is zero or one.</li>
     * </ul>
     *
     * @param array array to be copied.
     * @param offset specifies the first element of this array's range.
     * @param length specifies the range length of this array.
     * @return immutable array of byte values.
     * @throws IndexOutOfBoundsException if offset or length is invalid.
     */
    static ByteArray copyOf(@Nonnull byte[] array, int offset, int length) {
        return Arrays.newArray(array, offset, length, 0, true);
    }

    /**
     * Creates an immutable array representation of the specified {@code array}. The specified {@code array} is
     * used as backing array. It is therefore advised not to modify {@code array} after this immutable array is
     * created, as this would violate this array's immutability.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code NONNULL};</li>
     *   <li>{@code SORTED} and {@code DISTINCT} if the specified {@code array} contains zero or one element(s).</li>
     * </ul>
     *
     * @param array array to be wrapped.
     * @return immutable array that is backed by the specified {@code array}.
     */
    static ByteArray unsafeValueOf(@Nonnull byte... array) {
        return unsafeValueOf(array, 0, array.length);
    }

    /**
     * Creates an immutable array representation of the specified {@code array}. The specified {@code array} is
     * used as backing array. It is therefore advised not to modify {@code array} after this immutable array is
     * created, as this would violate this array's immutability.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code NONNULL};</li>
     *   <li>{@code SORTED} and {@code DISTINCT} if the specified {@code length} is zero or one.</li>
     * </ul>
     *
     * @param array array to be wrapped.
     * @param offset specifies the first element of this array's range.
     * @param length specifies the range length of this array.
     * @return immutable array that is backed by the specified {@code array}.
     */
    static ByteArray unsafeValueOf(@Nonnull byte[] array, int offset, int length) {
        return Arrays.newArray(array, offset, length, 0, false);
    }

    /**
     * Returns the byte value found at {@code [offset + index]} of the backing array.
     *
     * @param index index relative to the offset of the array.
     * @return byte value found at the specified index.
     * @throws IndexOutOfBoundsException if {@code index} position is invalid.
     */
    byte getByte(int index);

    /**
     * Returns {@code true} if this array contains the specified value.
     *
     * <p>The Java Array Framework implementation performs a binary search if this array reports {@code SORTED}
     * characteristics, otherwise all elements are traversed until the specified value is found.</p>
     *
     * @param value value whose presence in this array is to be tested.
     * @return {@code true} if this array contains the specified value.
     */
    boolean containsByte(byte value);

    /**
     * Returns the index of the first occurrence of the specified value in the array, or {@code -1} if
     * the array does not contain the value. The index is relative to the offset of the array.
     *
     * <p>The Java Array Framework implementation performs a binary search if this array reports {@code SORTED}
     * characteristics, otherwise all elements are traversed until the specified value is found.</p>
     *
     * @param value value whose lowest index in this array is to be returned.
     * @return lowest index of value in array or {@code -1} if value is not found.
     */
    int indexOfByte(byte value);

    /**
     * Returns the index of the last occurrence of the specified value in the array, or {@code -1} if
     * the array does not contain the value. The index is relative to the offset of the array.
     *
     * <p>The Java Array Framework implementation performs a binary search if this array reports {@code SORTED}
     * characteristics, otherwise all elements are traversed backwards until the specified value is found.</p>
     *
     * @param value value whose highest index in this array is to be returned.
     * @return highest index of value in array or {@code -1} if value is not found.
     */
    int lastIndexOfByte(byte value);

    /**
     * Iterates through the array and invokes the consumer's {@code accept} method for each value within array.
     *
     * @param consumer consumer that is called for each value within array's range.
     */
    void forEachByte(@Nonnull ByteConsumer consumer);

    /**
     * Returns an native array of the elements specified by {@code this} array. This method always returns a defensive
     * (range) copy of the backing array, even if this array is mutable.
     *
     * @return range copy of backing array.
     */
    @Override
    byte[] toArray();

    /**
     * Returns an array containing all of the byte values of {@code this} array. If the byte values fit in the
     * specified array, it is returned therein. Otherwise, a new byte array is allocated.
     *
     * @param array the array into which the elements of this array are to be stored, if it is big enough; otherwise, a
     *              new array of the same runtime type is allocated for this purpose.
     * @return an array containing all of the elements in the range specified by this array.
     * @throws NullPointerException if the specified array is null.
     */
    byte[] toArray(@Nonnull byte[] array);

    /**
     * <p>Returns a new sub array object for the specified {@code offset}. The specified {@code offset} is relative to
     * the existing offset, as returned by {@link #offset()}. The length of the resulting array is reduced by
     * the number of elements that are added to the current offset:</p>
     *
     * <pre>int newLength = length() - offset;</pre>
     *
     * <p>Effectively, the virtual {@code toIndex} of the array remains the same.</p>
     *
     * <p>A defensive copy of the backing array is created for immutable implementations. Mutable implementations
     * share the same backing array. The sub array has the same mutability properties as {@code this} array.</p>
     *
     * @param offset new relative offset.
     * @return a new sub array object with the new {@code offset}.
     */
    @Override
    ByteArray offset(int offset);    // More specific return type.

    /**
     * <p>Returns a new sub array object for the specified {@code length}. Note that {@code length} must be smaller than
     * or equal to the current length, as returned by {@code #length}.</p>
     *
     * <p>A defensive copy of the backing array is created for immutable implementations. Mutable implementations
     * share the same backing array.</p>
     *
     * @param length new length, must be smaller than or equal to current length.
     * @return a new sub array object with the new {@code length}.
     */
    @Override
    ByteArray length(int length);    // More specific return type.

    /**
     * Convenience method, combining {@code offset(int)} and {@code length(int)}.
     *
     * @param fromIndex low endpoint (inclusive) of the sub array.
     * @param toIndex high endpoint (exclusive) of the sub array.
     * @return a view of the specified range within this array.
     * @throws IndexOutOfBoundsException for an illegal endpoint index value ({@code fromIndex < 0 || toIndex > length ||
     *         fromIndex > toIndex}).
     */
    @Override
    ByteArray subArray(int fromIndex, int toIndex);  // More specific return type.

    /**
     * Returns an array iterator over the elements in this array.
     *
     * @return an array iterator over the elements in this array.
     */
    @Override
    ArrayIterator.OfByte iterator(); // More specific return type.

    /**
     * Returns an array iterator over the elements in this array.
     *
     * @param index index of the first element to be returned from the array iterator.
     * @return a array iterator over the elements in this array.
     * @throws IndexOutOfBoundsException if the index is out of range.
     */
    @Override
    ArrayIterator.OfByte iterator(int index);   // More specific return type.
}
