/*
 * ============================================================================
 * GNU Lesser General Public License
 * ============================================================================
 *
 * Square Brackets - Java Array Framework.
 * Copyright (C) 2013  Beatgrid Media B.V.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * Beatgrid Media B.V.
 * Sisalbaan 5-A
 * 2352 AZ Leiderdorp
 * The Netherlands
 *
 * info@squarebrackets.org
 * http://squarebrackets.org
 */
package org.squarebrackets;

import javax.annotation.Nonnull;
import java.util.Objects;
import java.util.function.Predicate;

/**
 * <p>Base interface for dynamic arrays.</p>
 *
 * <p>Arrays are constructed by the static interface methods that are declared by the typed interfaces of the array
 * hierarchy. Each of these static methods provide different semantics, such as {@code newInstance} for constructing
 * empty dynamic-size arrays, {@code valueOf} for wrapping {@code Array} around native arrays and {@code copyOf} for
 * copying the contents of native arrays or framework {@code Array}s.
 *
 * <p>Arrays express characteristics to indicate that, for instance, elements are listed in sorted order, or no
 * duplicate elements exist. These characteristics allow array implementations to optimize their algorithms based
 * on the contents of the array. For example, operations that lookup elements can use a binary search algorithm if
 * the array is marked as being sorted. These optimizations may reduce the time complexity of such operations by
 * several orders of magnitude.
 * Mutable arrays always report {@code MUTABLE}.</p>
 *
 * <p>The array implementations of the framework attempt to establish these characteristics to accommodate these
 * optimizations at best effort. This means that the array implementation determines which set of characteristics apply
 * to its elements without traversing these elements, in other words, without incurring a performance penalty.
 * Characteristics are always invalidated if conditions of these characteristics are broken as result of array
 * modification. Characteristics are only added in scenarios in which their conditions can be 'tested' without
 * traversal of elements. For example, the {@code SORTED} characteristic is added if the length of an array is reduced
 * to zero or one element, or if all elements are replaced by an array that is marked as being sorted.
 * The semantics of characteristics processing is documented for each "destructive" method. All implementations returned
 * by the framework adhere to these semantics. Custom implementations may choose to set characteristics at their
 * discretion as long as they do not violate the contract of these characteristics.
 * Users can take full control of characteristics assignment by means of the {@link MutableArray#doAction(ArrayAction)}
 * method.</p>
 *
 * @param <E> object type.
 * @author Leon van Zantvoort
 */
public interface DynamicArray<E> extends MutableArray<E> {

    /**
     * Specifies if reserved elements need to be reset to default values. This flag is set to {@code false} by default
     * for primitive arrays.
     *
     * @param flag {@code true} if reserved elements need to be reset, {@code false} otherwise.
     */
    void setClearReservedElements(boolean flag);

    /**
     * Appends the specified object to the end of this array's range. Length of {@code array} is increased by one.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect this array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>if this array reports {@code NONNULL} prior to this call, {@code NONNULL} is reported after this
     *   invocation if the specified {@code object} is not {@code null};</li>
     *   <li>if this array reports {@code SORTED} prior to this call, {@code SORTED} is reported after this
     *   invocation if the specified {@code object} does not break the sorted order of this array. Arrays
     *   sorted according to natural order are not regarded sorted if they contain {@code null} element(s);</li>
     *   <li>if this array reports {@code SORTED} and {@code DISTINCT} prior to this call, {@code DISTINCT} is
     *   reported after this invocation if the specified {@code object} was not already present in this array.</li>
     * </ul>
     *
     * @param object object to to be added.
     * @throws NullPointerException if object is null and array does not permit null elements.
     */
    void add(E object);

    /**
     * Inserts the specified object at the specified {@code index}, moving elements at and after {@code index} to the
     * right. Length of {@code this} array is increased by one.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect this array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>if this array reports {@code NONNULL} prior to this call, {@code NONNULL} is reported after this
     *   invocation if the specified {@code object} is not {@code null};</li>
     *   <li>if this array reports {@code SORTED} prior to this call, {@code SORTED} is reported after this
     *   invocation if the specified {@code object} does not break the sorted order of this array. Arrays
     *   sorted according to natural order are not regarded sorted if they contain {@code null} element(s);</li>
     *   <li>if this array reports {@code SORTED} and {@code DISTINCT} prior to this call, {@code DISTINCT} is
     *   reported after this invocation if the specified {@code object} was not already present in this array.</li>
     * </ul>
     *
     * @param index insertion index, relative to the offset of the array.
     * @param object object to be inserted.
     * @throws IndexOutOfBoundsException if {@code index} position is invalid.
     * @throws NullPointerException if object is null and array does not permit null elements.
     */
    void add(int index, E object);

    /**
     * Appends the elements of the specified array to the end of this array. Length of {@code this} array is increased
     * by {@code array.length()}.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>if this array reports {@code NONNULL} prior to this call, {@code NONNULL} is reported after this
     *   invocation if the specified {@code array} reports {@code NONNULL};</li>
     *   <li>if this array reports {@code SORTED} prior to this call, {@code SORTED} is reported after this
     *   invocation if the specified {@code array} reports {@code SORTED} and if it does not break the sorted order of
     *   this array;</li>
     *   <li>if this array reports {@code SORTED} and {@code DISTINCT} prior to this call, {@code DISTINCT} is reported
     *   after this invocation if the specified {@code array} reports {@code SORTED} and {@code DISTINCT} and if the
     *   specified {@code array} does not contain any objects that were already present in this array.</li>
     * </ul>
     *
     * @param array array to be added.
     * @throws NullPointerException if the specified array is null or if the array contains null elements and this
     * array does not permit null elements.
     */
    void addAll(@Nonnull Array<? extends E> array);

    /**
     * Inserts the elements of the specified array at the specified {@code index}, moving elements at and after
     * {@code index} to the right. Length of {@code this} array is increased by {@code array.length()}.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>if this array reports {@code NONNULL} prior to this call, {@code NONNULL} is reported after this
     *   invocation if the specified {@code array} reports {@code NONNULL};</li>
     *   <li>if this array reports {@code SORTED} prior to this call, {@code SORTED} is reported after this
     *   invocation if the specified {@code array} reports {@code SORTED} and if it does not break the sorted order of
     *   this array;</li>
     *   <li>if this array reports {@code SORTED} and {@code DISTINCT} prior to this call, {@code DISTINCT} is reported
     *   after this invocation if the specified {@code array} reports {@code SORTED} and {@code DISTINCT} and if the
     *   specified {@code array} does not contain any objects that were already present in this array.</li>
     * </ul>
     *
     * @param index insertion index, relative to the offset of the array.
     * @param array array to be inserted.
     * @throws IndexOutOfBoundsException if {@code index} position is invalid.
     * @throws NullPointerException if the specified array is null or if the array contains null elements and this
     * array does not permit null elements.
     */
    void addAll(int index, @Nonnull Array<? extends E> array);

    /**
     * Replaces all elements of this array by the elements of the specified array. The semantics of this method are less
     * strict than the ones expressed by {@code MutableArray}; The array that is passed as argument does not need to
     * have the same length as {@code this} array. This method does not throw an {@code IllegalArgumentException}.
     *
     * <p>The length of this array is increased or decreased by length difference between this array and the specified
     * {@code array}.</p>
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>if this array reports {@code NONNULL} prior to this call, {@code NONNULL} is reported after this
     *   invocation if the specified {@code array} reports {@code NONNULL};</li>
     *   <li>if this array reports {@code SORTED} prior to this call, {@code SORTED} is reported after this
     *   invocation if the specified {@code array} reports {@code SORTED} and if it does not break the sorted order of
     *   this array;</li>
     *   <li>if this array reports {@code SORTED} and {@code DISTINCT} prior to this call, {@code DISTINCT} is reported
     *   after this invocation if the specified {@code array} reports {@code SORTED} and {@code DISTINCT} and if the
     *   specified {@code array} does not contain any objects that were already present in this array.</li>
     * </ul>
     *
     * @param array array to replace elements of this array.
     * @throws NullPointerException if the specified array is null or if the element is replaced with a null value
     * and this array does not permit null elements.
     */
    @Override
    void setAll(@Nonnull Array<? extends E> array);

    /**
     * Removes element at specified {@code index}, moving all elements after {@code index} to the left. Length of
     * {@code this} array is decreased by one.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>if the array contains no elements after this invocation, {@code NONNULL}, {@code SORTED} and
     *   {@code DISTINCT} are reported.
     * </ul>
     *
     * @param index removal index, relative to the offset of the array.
     * @return element that is removed.
     * @throws IndexOutOfBoundsException if {@code index} position is invalid.
     */
    E remove(int index);

    /**
     * Removes the first occurrence of the specified object from this array if it is present. If this array does not
     * contain the object, it is unchanged. This method supports boxed primitives. This means that if this array
     * represents a primitive array, this method removes the first primitive element that is the unboxed counterpart of
     * the object.
     *
     * <p>The Java Array Framework implementation performs a binary search if this array reports {@code SORTED}
     * characteristics to lookup the element to be removed, otherwise all elements are traversed until the specified
     * object is found.</p>
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>if the array contains no elements after this invocation, {@code NONNULL}, {@code SORTED} and
     *   {@code DISTINCT} are reported.
     * </ul>
     *
     * @param object object to be removed.
     * @return {@code true} if this array contained the specified object.
     * @throws NullPointerException if object is null and array is sorted by a comparator that does not permit null arguments.
     * @throws ClassCastException if this array is sorted and comparator not compatible with object.
     */
    boolean remove(Object object);

    /**
     * Removes all of this array's elements that are also contained in the specified array. Primitives and their object
     * counterparts are regarded equal in this respect. After this call returns, this array will contain no elements
     * in common with the specified array.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>if the array contains no elements after this invocation, {@code NONNULL}, {@code SORTED} and
     *   {@code DISTINCT} are reported.
     * </ul>
     *
     * @param array array containing elements to be removed from this array.
     * @return {@code true} if this array changed as a result of the call.
     * @throws NullPointerException if the specified array is null. Exception may also be thrown if this or specified
     * array contain null elements.
     * @throws ClassCastException if this or specified array is sorted and an element of this array is incompatible with
     * the specified array.
     */
    boolean removeAll(@Nonnull Array<?> array);

    /**
     * Retains only the elements in this array that are contained in the specified array. In other words, removes from
     * this array all of its elements that are not contained in the specified array. Primitives and their object
     * counterparts are regarded equal in this respect.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>if the array contains no elements after this invocation, {@code NONNULL}, {@code SORTED} and
     *   {@code DISTINCT} are reported.
     * </ul>
     *
     * @param array array containing elements to be retained in this array.
     * @return {@code true} if this array changed as a result of the call.
     * @throws NullPointerException if the specified array is null. Exception may also be thrown if this or specified
     * array contain null elements.
     * @throws ClassCastException if this or specified array is sorted and an element of this array is incompatible with
     * the specified array.
     */
    boolean retainAll(@Nonnull Array<?> array);

    /**
     * Removes all elements of range, moving elements from the right after range to offset position.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>After this invocation, {@code NONNULL}, {@code SORTED} and {@code DISTINCT} are reported.
     * </ul>
     */
    void clear();

    /**
     * Replaces backing array by a new array that has the same length as the root array, if and only if the
     * backing is greater than the root array.
     */
    void compact();

    /**
     * Removes all of the elements of this array's range that satisfy the given predicate. Errors or runtime exceptions
     * thrown by the predicate are relayed to the caller.
     *
     * <p>The default implementation traverses all elements of the array using its {@link #iterator}. Each matching
     * element is removed using {@link java.util.Iterator#remove()}.</p>
     *
     * @param filter a predicate which returns {@code true} for elements to be removed.
     * @return {@code true} if any elements were removed.
     * @throws NullPointerException if the specified filter is null.
     */
    default boolean removeIf(@Nonnull Predicate<? super E> filter) {
        Objects.requireNonNull(filter);
        boolean removed = false;
        final ArrayIterator<E> each = iterator();
        while (each.hasNext()) {
            if (filter.test(each.next())) {
                each.remove();
                removed = true;
            }
        }
        return removed;
    }

    /**
     * Removes all duplicates from this array. The order of elements may be altered as result of this call.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>{@code SORTED} and {@code DISTINCT} are reported after this invocation.</li>
     * </ul>
     *
     * @throws NullPointerException if this array contains null elements.
     * @throws ClassCastException if this array contains elements that are not mutually {@code Comparable}.
     */
    void distinct();

    /**
     * <p>Returns a new sub array object for the specified {@code offset}. The specified {@code offset} is relative to
     * the existing offset, as returned by {@link #offset()}. The length of the resulting array is reduced by
     * the number of elements that are added to the current offset:</p>
     *
     * <pre>int newLength = length() - offset;</pre>
     *
     * <p>Effectively, the virtual {@code toIndex} of the array remains the same.</p>
     *
     * <p>A defensive copy of the backing array is created for immutable implementations. Mutable implementations
     * share the same backing array. The sub array has the same mutability properties as {@code this} array.</p>
     *
     * @param offset new relative offset.
     * @return a new sub array object with the new {@code offset}.
     */
    @Override
    DynamicArray<E> offset(int offset);

    /**
     * <p>Returns a new sub array object for the specified {@code length}. Note that {@code length} must be smaller than
     * or equal to the current length, as returned by {@code #length}.</p>
     *
     * <p>A defensive copy of the backing array is created for immutable implementations. Mutable implementations
     * share the same backing array.</p>
     *
     * @param length new length, must be smaller than or equal to current length.
     * @return a new sub array object with the new {@code length}.
     */
    @Override
    DynamicArray<E> length(int length);

    /**
     * Convenience method, combining {@code offset(int)} and {@code length(int)}.
     *
     * @param fromIndex low endpoint (inclusive) of the sub array.
     * @param toIndex high endpoint (exclusive) of the sub array.
     * @return a view of the specified range within this array.
     * @throws IndexOutOfBoundsException for an illegal endpoint index value ({@code fromIndex < 0 || toIndex > length ||
     *         fromIndex > toIndex}).
     */
    @Override
    DynamicArray<E> subArray(int fromIndex, int toIndex);
}
