/*
 * ============================================================================
 * GNU Lesser General Public License
 * ============================================================================
 *
 * Square Brackets - Java Array Framework.
 * Copyright (C) 2013  Beatgrid Media B.V.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * Beatgrid Media B.V.
 * Sisalbaan 5-A
 * 2352 AZ Leiderdorp
 * The Netherlands
 *
 * info@squarebrackets.org
 * http://squarebrackets.org
 */
package org.squarebrackets;

import javax.annotation.Nonnull;
import java.io.IOException;
import java.io.InvalidObjectException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.NoSuchElementException;

import static java.lang.System.arraycopy;

/**
 * Mutable array of boolean primitives.
 *
 * @author Leon van Zantvoort
 */
class MutableBooleanArrayImpl extends BooleanArrayImpl implements MutableBooleanArray, MutableArrayCharacteristics {

    /** Controls state of array. Is {@code null} if control is transferred to caller of {@code doAction}. */
    transient ArrayContext context;

    /**
     * Package visible constructor.
     *
     * @param array backing array.
     * @param offset start index of range.
     * @param length length of range.
     * @param characteristics array characteristics.
     * @param copy {@code true} to make a defensive copy of {@code array}, {@code false} to use {@code array} reference
     *             as backing array.
     */
    MutableBooleanArrayImpl(@Nonnull boolean[] array, int offset, int length, int characteristics, boolean copy) {
        super(array, offset, length, characteristics, copy);
        context = new ArrayContextImpl();
    }

    /**
     * Returns array context, or {@code null} if {@code doAction} is being executed.
     *
     * @return array context, or {@code null} if {@code doAction} is being executed.
     */
    protected ArrayContext getArrayContext() {
        return context;
    }

    /**
     * Updates the SORTED and DISTINCT characteristics of this array by looking at the elements surrounding the
     * specified {@code value}. The SORTED and/or DISTINCT characteristics are removed if the value invalidates the
     * characteristics.
     *
     * @param elementIndex index relative to the backing array.
     * @param value value that is stored at index.
     */
    protected void updateCharacteristics(int elementIndex, boolean value) {
        assert hasCharacteristics(SORTED);
        if (length > 1) {
            if (elementIndex > offset) {
                if (!value && elements[elementIndex - 1]) {
                    removeCharacteristics(SORTED | DISTINCT);
                } else if (value == elements[elementIndex - 1]) {
                    removeCharacteristics(DISTINCT);
                }
            }
            if (elementIndex < offset + length - 1) {
                if (value && !elements[elementIndex + 1]) {
                    removeCharacteristics(SORTED | DISTINCT);
                } else if (value == elements[elementIndex + 1]) {
                    removeCharacteristics(DISTINCT);
                }
            }
        }
    }

    /**
     * Updates the SORTED and DISTINCT characteristics of this array by looking at the elements surrounding the
     * first and last element of the specified {@code array}. The SORTED and/or DISTINCT characteristics are removed if
     * these elements invalidate the characteristics.
     *
     * @param elementIndex index relative to the backing array.
     * @param array array that is stored, starting at index.
     */
    protected void updateCharacteristics(int elementIndex, @Nonnull Array<? extends Boolean> array) {
        assert hasCharacteristics(SORTED);
        int len = array.length();
        if (length > len) {
            if (len > 0) {
                updateCharacteristics(elementIndex, elements[elementIndex]);
            }
            if (len > 1) {
                updateCharacteristics(elementIndex + len - 1, elements[elementIndex + len - 1]);
            }
        }
    }

    /**
     * Sets the specified characteristics for this array.
     *
     * @param characteristics characteristics to be set for this array.
     * @return the characteristics that were actually stored for this array.
     */
    @Override
    public int characteristics(int characteristics) {
        if (characteristics != this.characteristics) { // Optimization.
            ArrayContext context = getArrayContext();
            if (context != null) {
                return context.characteristics(characteristics);
            }
        }
        return this.characteristics;
    }

    /**
     * Allow array implementation to filter characteristics based on array specific logic.
     *
     * <p>Default implementation passes specified value straight through.</p>
     *
     * @param characteristics characteristics to be filtered.
     * @return the filtered characteristics.
     */
    @Override
    protected int filterCharacteristics(int characteristics) {
        return characteristics | NONNULL | MUTABLE & ~DYNAMIC;
    }

    /**
     * Performs the specified {@code action} without affecting the array's set of characteristics. Ownership of
     * array characteristics control is moved to the caller by means of the {@code ArrayContext} object. Errors or
     * runtime exceptions thrown by the action are relayed to the caller.
     *
     * @param <T> return type.
     * @param action the action to be performed.
     * @return the value returned by the action's {@code run} method.
     * @throws IllegalStateException if action is called when array does not own control of characteristics.
     * @exception NullPointerException if the action is {@code null}.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    @Override
    public final <T> T doAction(@Nonnull ArrayAction<T> action) {
        ArrayContext context = getArrayContext();
        if (context == null) {
            throw new IllegalStateException();
        }
        try {
            this.context = null;
            return action.run(context);
        } finally {
            this.context = context;
        }
    }

    /**
     * Returns the backing array (optional operation). All elements of {@code this} mutable array and parents of this
     * mutable array are stored in the returned object. Modifications to the backing array are reflected by {@code this}
     * mutable array and it's parents.
     *
     * <p>The resulting object must be cast to the primitive or object type of the implementing class.</p>
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>This method allows the caller to modify this array outside the scope of the implementing class. Hence, it can
     * not guarantee correctness of the characteristics of this array. For that reason, content related characteristics
     * (such as {@code SORTED} and {@code DISTINCT}) are removed by calling this method.
     * Characteristics removal can be suppressed by obtaining the backing array through {@code doAction}:</p>
     *
     * <pre>
     *{@literal boolean[] backingArray = (boolean[]) array.doAction(ArrayContext::array);}</pre>
     *
     * @return backing array.
     * @throws UnsupportedOperationException if array does not expose backing array.
     */
    @Override
    public boolean[] array() {
        removeCharacteristics(characteristics);
        return elements;
    }

    /**
     * Replaces element at specified {@code index} by specified object.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect this array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>if this array reports {@code SORTED} prior to this call, {@code SORTED} is reported after this
     *   invocation if the specified {@code object} does not break the sorted order of this array. Arrays
     *   sorted according to natural order are not regarded sorted if they contain {@code null} element(s);</li>
     *   <li>if this array reports {@code SORTED} and {@code DISTINCT} prior to this call, {@code DISTINCT} is
     *   reported after this invocation if the specified {@code object} was not already present in this array.</li>
     * </ul>
     *
     * @param index index of item to be replaced, relative to the offset of the array.
     * @param object object to replace existing value.
     * @return object that is replaced.
     * @throws IndexOutOfBoundsException if {@code index} position is invalid.
     * @throws NullPointerException if specified {@code object} is null.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    @Override
    public Boolean set(int index, Boolean object) {
        if (Tripwire.ENABLED) {
            Tripwire.trip(getClass());
        }
        return setBoolean(index, object);
    }

    /**
     * Replaces element at specified {@code index} by specified boolean value.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect this array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>if this array reports {@code SORTED} prior to this call, {@code SORTED} is reported after this
     *   invocation if the specified {@code object} does not break the sorted order of this array;</li>
     *   <li>if this array reports {@code SORTED} and {@code DISTINCT} prior to this call, {@code DISTINCT} is
     *   reported after this invocation if the specified {@code object} was not already present in this array.</li>
     * </ul>
     *
     * @param index index relative to the offset of the array.
     * @param value boolean value to replace existing value.
     * @return boolean value that is replaced.
     * @throws IndexOutOfBoundsException if {@code index} position is invalid.
     */
    @Override
    public boolean setBoolean(int index, boolean value) {
        rangeCheck(index);
        int elementIndex = offset + index;
        boolean oldValue = elements[elementIndex];
        elements[elementIndex] = value;
        if (hasCharacteristics(SORTED)) {
            updateCharacteristics(elementIndex, value);
        }
        return oldValue;
    }

    /**
     * Replaces all elements of this array by the elements of the specified array. Note that the length of {@code array}
     * needs to be equal to the length of {@code this} array, throws an {@code IllegalArgumentException} otherwise.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>if this array reports {@code SORTED} prior to this call, {@code SORTED} is reported after this
     *   invocation if the specified {@code array} reports {@code SORTED};</li>
     *   <li>if this array reports {@code SORTED} and {@code DISTINCT} prior to this call, {@code DISTINCT} is
     *   reported after this invocation if the specified {@code array} reports {@code SORTED} and {@code DISTINCT}.</li>
     * </ul>
     *
     * @param array array to replace elements of this array.
     * @throws IllegalArgumentException if length of specified array is not equal to length of this array.
     * @throws NullPointerException if the specified array is null or if the array contains null elements and this
     * array does not permit null elements.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    @Override
    public void setAll(@Nonnull Array<? extends Boolean> array) {
        int lengthDiff = array.length() - length;
        if (lengthDiff == 0) {
            arraycopy(UnsafeArray.booleanReadAccess(array), array.offset(), elements, offset, length);
            characteristics(array.characteristics());   // All elements are replaced by new array; Copy characteristics.
            if (hasCharacteristics(SORTED)) {
                updateCharacteristics(offset, array);
            }
        } else {
            throw new IllegalArgumentException(String.format("Array length (%d) not equal to length of this array (%d).",
                    array.length(), length));
        }
    }

    /**
     * Sorts this array according to natural ordering.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>{@code SORTED} is reported after this invocation.</li>
     * </ul>
     */
    @Override
    public void sort() {
        if (!hasCharacteristics(SORTED)) {
            Arrays.sort(elements, offset, offset + length);
            addCharacteristics(SORTED);  // Array is guaranteed to be sorted.
        }
    }

    /**
     * Sorts this array according to natural ordering. The underlying implementation utilizes multiple threads
     * to reduce sorting time. The algorithm may require a working space equal to the size of this array.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>{@code SORTED} is reported after this invocation.</li>
     * </ul>
     */
    @Override
    public void parallelSort() {
        if (!hasCharacteristics(SORTED)) {
            Arrays.parallelSort(elements, offset, offset + length);
            addCharacteristics(SORTED); // Array is guaranteed to be sorted.
        }
    }

    /**
     * Array context implementation.
     */
    @SuppressWarnings("MissingMethodJavaDoc")
    private class ArrayContextImpl implements ArrayContext {

        @Override
        public Object array() {
            return elements;
        }

        @Override
        public int characteristics(int newCharacteristics) {
            characteristics = filterCharacteristics(newCharacteristics);
            return characteristics;
        }

        @Override
        public int characteristics() {
            return characteristics;
        }
    }

    /**
     * Sub arrays share serialization proxy with parent arrays intentionally. As a result, deserialized sub arrays become
     * detached from their originating parent.
     */
    @SuppressWarnings({"MissingMethodJavaDoc", "MissingFieldJavaDoc"})
    private static class SubMutableBooleanArrayImpl extends MutableBooleanArrayImpl {

        final MutableBooleanArrayImpl parent;

        SubMutableBooleanArrayImpl(@Nonnull MutableBooleanArrayImpl parent, int offset, int length) {
            super(parent.elements, offset, length, parent.characteristics, false);
            this.parent = parent;
        }

        @Override
        protected ArrayContext getArrayContext() {
            return parent.getArrayContext() == null ? null : context;
        }

        @Override
        protected void updateCharacteristics(int elementIndex, boolean value) {
            super.updateCharacteristics(elementIndex, value);
            if (hasCharacteristics(SORTED)) {
                parent.updateCharacteristics(elementIndex, value);
            }
        }

        @Override
        protected void updateCharacteristics(int elementIndex, @Nonnull Array<? extends Boolean> array) {
            super.updateCharacteristics(elementIndex, array);
            if (hasCharacteristics(SORTED)) {
                parent.updateCharacteristics(elementIndex, array);
            }
        }

        @Override
        public int characteristics(int characteristics) {
            // Parent array retains characteristics that are shared with sub array. Additional characteristics
            // assigned to sub array are not set for parent array, as it is not guaranteed that these
            // characteristics hold for parent array.
            parent.retainCharacteristics(characteristics);
            return super.characteristics(characteristics);
        }
    }

    /**
     * <p>Returns a new sub array object for the specified {@code offset}. The specified {@code offset} is relative to
     * the existing offset, as returned by {@link #offset()}. The length of the resulting array is reduced by
     * the number of elements that are added to the current offset:</p>
     *
     * <pre>int newLength = length() - offset;</pre>
     *
     * <p>Effectively, the virtual {@code toIndex} of the array remains the same.</p>
     *
     * <p>A defensive copy of the backing array is created for immutable implementations. Mutable implementations
     * share the same backing array. The sub array has the same mutability properties as {@code this} array.</p>
     *
     * @param offset new relative offset.
     * @return a new sub array object with the new {@code offset}.
     */
    @Override
    public MutableBooleanArray offset(int offset) {
        if (offset == 0) {
            return this;
        }
        int newOffset = this.offset + offset;
        int newLength = length - offset;
        subArrayCheck(newOffset, newLength);
        return new SubMutableBooleanArrayImpl(this, newOffset, newLength);
    }

    /**
     * <p>Returns a new sub array object for the specified {@code length}. Note that {@code length} must be smaller than
     * or equal to the current length, as returned by {@code #length}.</p>
     *
     * <p>A defensive copy of the backing array is created for immutable implementations. Mutable implementations
     * share the same backing array.</p>
     *
     * @param length new length, must be smaller than or equal to current length.
     * @return a new sub array object with the new {@code length}.
     */
    @Override
    public MutableBooleanArray length(int length) {
        if (this.length == length) {
            return this;
        }
        subArrayCheck(offset, length);
        return new SubMutableBooleanArrayImpl(this, offset, length);
    }

    /**
     * Convenience method, combining {@code offset(int)} and {@code length(int)}.
     *
     * @param fromIndex low endpoint (inclusive) of the sub array.
     * @param toIndex high endpoint (exclusive) of the sub array.
     * @return a view of the specified range within this array.
     * @throws IndexOutOfBoundsException for an illegal endpoint index value ({@code fromIndex < 0 || toIndex > length ||
     *         fromIndex > toIndex}).
     */
    @Override
    public MutableBooleanArray subArray(int fromIndex, int toIndex) {
        int newOffset = offset + fromIndex;
        int newLength = toIndex - fromIndex;
        if (newOffset == offset && newLength == length) {
            return this;
        }
        subArrayCheck(newOffset, newLength);
        return new SubMutableBooleanArrayImpl(this, newOffset, newLength);
    }

    @SuppressWarnings({"MissingMethodJavaDoc", "MissingFieldJavaDoc"})
    private class ArrayIteratorImpl implements ArrayIterator.OfBoolean {
        int index;
        int lastIndex = -1;

        ArrayIteratorImpl() {
            this(0);
        }

        ArrayIteratorImpl(int index) {
            this.index = offset + index;
        }

        @Override
        public boolean hasNext() {
            return index < offset + length;
        }

        @Override
        public boolean nextBoolean() {
            int i = index;
            if (i >= offset + length) {
                throw new NoSuchElementException();
            }
            assert i < elements.length;
            index = i + 1;
            return elements[lastIndex = i];
        }

        @Override
        public void remove() {
            throw new UnsupportedOperationException();
        }

        @Override
        public boolean hasPrevious() {
            return index > offset;
        }

        @Override
        public boolean previousBoolean() {
            int i = index - 1;
            if (i < offset) {
                throw new NoSuchElementException();
            }
            assert i < elements.length;
            index = i;
            return elements[lastIndex = i];
        }

        @Override
        public int nextIndex() {
            return index;
        }

        @Override
        public int previousIndex() {
            return index - 1;
        }

        @Override
        public void setBoolean(boolean value) {
            if (lastIndex < 0) {
                throw new IllegalStateException();
            }
            assert lastIndex < offset + length && lastIndex < elements.length;
            MutableBooleanArrayImpl.this.set(lastIndex, value);
        }

        @Override
        public void addBoolean(boolean value) {
            throw new UnsupportedOperationException();
        }
    }

    /**
     * Returns an array iterator over the elements in this array.
     *
     * @return an array iterator over the elements in this array.
     */
    @Override
    public ArrayIterator.OfBoolean iterator() {
        return new ArrayIteratorImpl();
    }

    /**
     * Returns an array iterator over the elements in this array.
     *
     * @param index index of the first element to be returned from the array iterator.
     * @return a array iterator over the elements in this array.
     * @throws IndexOutOfBoundsException if the index is out of range.
     */
    @Override
    public ArrayIterator.OfBoolean iterator(int index) {
        if (index > length) {
            throw new IndexOutOfBoundsException();
        }
        return new ArrayIteratorImpl(index);
    }

    /**
     * Write replace method for the serialization proxy pattern. This method must be overwritten by sub classes if their
     * serialized form differs from this class.
     *
     * @return serialization proxy.
     */
    @Override
    Object writeReplace() {
        return new SerializationProxy(this);
    }

    @SuppressWarnings("MissingMethodJavaDoc")
    private void readObject(ObjectInputStream s) throws InvalidObjectException {
        throw new InvalidObjectException("Proxy required");
    }

    @SuppressWarnings({"MissingMethodJavaDoc", "MissingFieldJavaDoc"})
    private static class SerializationProxy implements Serializable {
        static final long serialVersionUID = -4560912380643845403L;

        transient int offset;
        final int length;
        final int characteristics;
        transient boolean[] elements;

        SerializationProxy(MutableBooleanArrayImpl array) {
            this.offset = array.offset;
            this.length = array.length;
            this.characteristics = array.characteristics;
            this.elements = array.elements;
        }

        private void writeObject(ObjectOutputStream s) throws IOException {
            s.defaultWriteObject();
            for (int i = 0; i < length; i++) {
                s.writeBoolean(elements[offset + i]);
            }
        }

        private void readObject(ObjectInputStream s) throws IOException, ClassNotFoundException {
            s.defaultReadObject();
            elements = new boolean[length];
            for (int i = 0; i < length; i++) {
                elements[i] = s.readBoolean();
            }
        }

        private Object readResolve() {
            return new MutableBooleanArrayImpl(elements, 0, length, characteristics, false);
        }
    }
}
