/*
 * ============================================================================
 * GNU Lesser General Public License
 * ============================================================================
 *
 * Square Brackets - Java Array Framework.
 * Copyright (C) 2013  Beatgrid Media B.V.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * Beatgrid Media B.V.
 * Sisalbaan 5-A
 * 2352 AZ Leiderdorp
 * The Netherlands
 *
 * info@squarebrackets.org
 * http://squarebrackets.org
 */
package org.squarebrackets.function;

import java.util.Objects;

/**
 * An operation on a single {@code short} operand yielding an {@code short} result.
 */
@FunctionalInterface
public interface ShortUnaryOperator {

    /**
     * Returns the {@code short} value result of the operation upon the
     * {@code short} operand.
     *
     * @param operand the operand value
     * @return the operation result value
     */
    short applyAsShort(short operand);

    /**
     * Compose a new function which applies the provided function followed by this function. If either function throws
     * an exception, it is relayed to the caller.
     *
     * @param before an additional function to be applied before this function is applied.
     * @return a function which performs the provided function followed by this function.
     * @throws NullPointerException if before is null.
     */
    default ShortUnaryOperator compose(ShortUnaryOperator before) {
        Objects.requireNonNull(before);
        return (short v) -> applyAsShort(before.applyAsShort(v));
    }

    /**
     * Compose a new function which applies this function followed by the provided function. If either function throws
     * an exception, it is relayed to the caller.
     *
     * @param after an additional function to be applied after this function is applied.
     * @return a function which performs this function followed by the provided function followed.
     * @throws NullPointerException if after is null.
     */
    default ShortUnaryOperator andThen(ShortUnaryOperator after) {
        Objects.requireNonNull(after);
        return (short t) -> after.applyAsShort(applyAsShort(t));
    }

    /**
     * Returns a unary operator that provides its input value as the result.
     *
     * @return a unary operator that provides its input value as the result.
     */
    static ShortUnaryOperator identity() {
        return t -> t;
    }
}
